{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE QuantifiedConstraints #-}
{-# OPTIONS_GHC -Wno-orphans #-}
module Types where

import Data.Aeson hiding (Options)
import Data.Default
import Data.StateVar
import Data.Types.Injective
import Data.Time
import Data.FingerTree
import Data.Semigroup
import Generic.Data.Microsurgery
import GHC.Generics
import SDL (Window)
import DearImGui
import RIO
import RIO.Process
import Lens.Micro.TH
import qualified Data.Text as T
import qualified Data.Vector.Storable as VS

import IBClient.Types

-- | Command line arguments
data Options = Options
  { optionsVerbose :: !Bool
  }

data WindowParams = WindowParams
  { _windowWidth :: Int
  , _windowHeight :: Int
  } deriving stock (Show, Generic)
    deriving anyclass (FromJSON, ToJSON)

instance Default WindowParams where
  def = WindowParams 1024 768

data TWSConnection = TWSConnection
  { _host :: Text
  , _port :: Text
  } deriving stock (Show, Generic)
    deriving anyclass (FromJSON, ToJSON)

instance Default TWSConnection where
  def = TWSConnection "127.0.0.1" "7497"

instance FromJSON LogLevel where
  parseJSON = withText "LogLevel" $ \case
    "LevelDebug"  -> return LevelDebug
    "LevelInfo"   -> return LevelInfo
    "LevelWarn"   -> return LevelWarn
    "LevelError"  -> return LevelError
    x -> fail $ T.unpack $ "encountered "<>x

instance ToJSON LogLevel where
  toJSON  LevelDebug    = String "LevelDebug"
  toJSON  LevelInfo     = String "LevelInfo"
  toJSON  LevelWarn     = String "LevelWarn"
  toJSON  LevelError    = String "LevelError"
  toJSON (LevelOther _) = String "LevelDebug"

data Settings = Settings
  { _windowParams  :: WindowParams
  , _twsConnection :: TWSConnection
  , _logLevel      :: LogLevel
  } deriving stock (Show, Generic)
    deriving anyclass (FromJSON, ToJSON)

makeLenses ''WindowParams
makeLenses ''TWSConnection
makeLenses ''Settings

instance Default Settings where
  def = Settings def def LevelWarn

data TWSConnectionStatus = TWSDisconnected
                         | TWSConnecting
                         | TWSConnected
                         deriving stock (Show, Eq, Enum, Bounded)

data TWSConnectionRefs = TWSConnectionRefs
  { twsConnectionRefsHost :: TVar Text
  , twsConnectionRefsPort :: TVar Text
  , twsConnectionStatus   :: TVar TWSConnectionStatus
  , twsConnectionSend     :: TQueue Msg_IB_OUT
  , twsConnectionRecieve  :: TQueue Msg_IB_IN
  }

instance Injective TWSConnectionStatus ImVec4 where
  to = \case
      TWSDisconnected -> ImVec4 1 0 0 1
      TWSConnecting   -> ImVec4 1 1 0 1
      TWSConnected    -> ImVec4 0 1 0 1

instance Injective TWSConnectionStatus Text where
  to = \case
      TWSDisconnected -> "Not Connected"
      TWSConnecting   -> "Trying to connect..."
      TWSConnected    -> "Connected"

data DataRefs = DataRefs
              { accounts            :: TVar (HashMap Text IBAccount)
              , nextValidID         :: TVar (Maybe Int)
              , nextSymbolLookup    :: TVar Text
              , symbolLookupResults :: TVar [IBSymbolSample]
              }
mkIBAccount :: Text -> IBAccount
mkIBAccount u = IBAccount (IBAccountInfo u mempty mempty) mempty mempty

data IBAccount = IBAccount
               { _accountInfo       :: IBAccountInfo
               , _accountPortfolio  :: [IBPortfolioValue]
               , _accountStrategies :: [IBAccountStrategy]
               } deriving stock (Show, Eq)


data IBAccountInfo = IBAccountInfo
                   { _accountName       :: Text
                   , _accountProperties :: HashMap Text [(Text, Text)] -- (value, currency)
                   , _accountLastUpdate :: Text
                   } deriving stock (Show, Eq)

data IBPortfolioValue = IBPortfolioValue
                      { _contract      :: IBContract
                      , _position      :: Float
                      , _marketPrice   :: Float
                      , _marketValue   :: Float
                      , _averageCost   :: Float
                      , _unrealizedPNL :: Float
                      , _realizedPNL   :: Float
                      } deriving stock (Show, Eq)

newtype IBAccountStrategy = IBAccountStrategy [Int] -- Int -> contract-id for this strategy
  deriving stock (Show)
  deriving newtype (Eq)

data IBSymbolSample = IBSymbolSample
                    { _symbolId :: Int
                    , _symbol :: Text
                    , _secType :: Text
                    , _primaryExchange :: Text
                    , _currency :: Text
                    , _derivatives :: [Text]
                    } deriving stock (Show, Eq)

makeLenses ''IBAccountStrategy
makeLenses ''IBAccountInfo
makeLenses ''IBAccount

data ChartSettings = ChartSettings
                   { chartResolution    :: Int
                   , chartStart         :: Maybe UTCTime
                   , chartEnd           :: Maybe UTCTime
                   , chartStudySettings :: [ChartStudyType]
                   } deriving stock (Show, Eq)

defChartSettings :: ChartSettings
defChartSettings = ChartSettings 60 Nothing Nothing [ChartStudyTypeDirect]

updateChartStudySettings :: Chart -> [ChartStudyType] -> Chart
updateChartStudySettings Chart{..} s = 
    let chartSettings   = chartSettings { chartStudySettings = s }
        chartCache      = emptyChartCacheData chartCacheSettings
        lastCacheUpdate = Nothing
    in Chart{..}

-- | Settings for Chart-cache.
--
-- History and Resolution yield 1 day by default. Do not increase over 5000 as memery-consumption explodes!
-- Think of it as timePlotted = candleWidth * numberOfCandles
--
-- chartCacheHistory should be a multiple of 20 as there often is 5% buffer in various caches and this
-- fills the caches more optimally
data ChartCacheSettings = ChartCacheSettings
                        { chartCacheResolution :: Int
                        , chartCacheHistory    :: Int
                        } deriving stock (Show, Eq)

defChartCacheSettings :: ChartCacheSettings
--defChartCacheSettings = ChartCacheSettings 60 1440
defChartCacheSettings = ChartCacheSettings 5 20

-- TODO: TimePointFloat? or only 1 entry per second?

data TimeInterval = TimeInterval
                  { timeIntervalBegin :: Int
                  , timeIntervalEnd   :: Int
                  }
  deriving stock (Generic, Eq)
  deriving anyclass (FromJSON, ToJSON) -- TODO: write own instances with "TimeInterval [a,b]" instead of recods.
  deriving (Show) via (Surgery Derecordify TimeInterval)
instance Semigroup TimeInterval where
  (TimeInterval mi1 ma1) <> (TimeInterval mi2 ma2)
    = TimeInterval (getMin $ Min mi1 <> Min mi2) (getMax $ Max ma1 <> Max ma2)
instance Monoid TimeInterval where
  mempty = TimeInterval (getMin mempty) (getMax mempty)


newtype TimePoint = TimePoint Int
   deriving stock (Generic, Show)
   deriving newtype (Eq, Enum, Real, Ord, Num, Integral, FromJSON, ToJSON)
   deriving (Semigroup, Monoid) via (Min Int)

intervalFrom :: TimeInterval -> TimePoint
intervalFrom (TimeInterval a _) = TimePoint a

intervalTo :: TimeInterval -> TimePoint
intervalTo (TimeInterval _ b) = TimePoint b

data ChartStudies = SMA { window :: Int, value :: Float }
                  | OLHC { olhc_open :: Float, olhc_low :: Float, olhc_high :: Float, olhc_close :: Float}
                  | Volume { volume :: Float }
                  deriving stock (Show, Eq, Generic)
                  deriving anyclass (FromJSON, ToJSON)

data ChartPoint = ChartPoint
                { timeOfDay   :: TimeInterval
                , pointValue  :: Float
                , pointVolume :: Float
                , pointExtra  :: [ChartStudies]
                } deriving stock (Show, Eq, Generic)
                  deriving anyclass (FromJSON, ToJSON)

instance Measured TimeInterval ChartPoint where
  measure = timeOfDay

-- | Tick-based data
data ChartStudyType = ChartStudyTypeDirect
                    | ChartStudyTypeSMA Int
                    | ChartStudyTypeOpen
                    | ChartStudyTypeLow
                    | ChartStudyTypeHigh
                    | ChartStudyTypeClose
                    | ChartStudyTypeVolume
                    deriving stock (Show, Eq, Generic)
                    deriving anyclass (Hashable)

data ChartCacheData = ChartCacheData
                    { chartCacheData     :: HashMap ChartStudyType (VS.Vector Float)
                    , chartCacheAxis     :: VS.Vector Float
                    , chartCacheCurrent  :: Maybe TimeInterval -- ^ in number of TimeInterval, negative meaning "before today"
                    , chartCacheFilledTo :: Maybe (Int,Int) -- ^ in index into the vectors given range of "Current"
                    } deriving stock (Show, Eq)

emptyChartCacheData :: ChartCacheSettings -> ChartCacheData
emptyChartCacheData ChartCacheSettings{..} = ChartCacheData mempty (VS.replicate chartCacheHistory 0) Nothing Nothing

data Chart = Chart
           { chartContractID    :: Int
           , chartData          :: FingerTree TimeInterval ChartPoint -- ^ raw data (time & sale)
           , chartHistData      :: HashMap Day (FingerTree TimeInterval ChartPoint) -- ^ raw data (time & sale)
           , fillerThread       :: ThreadId
           , chartSettings      :: ChartSettings
           , chartCacheSettings :: ChartCacheSettings
           , chartCache         :: ChartCacheData -- ^ cache for drawing with all pointExtra filled out and sampled according to 'chartSettings'
           , lastCacheUpdate    :: Maybe TimePoint
           , chartDirty         :: Bool
           } deriving stock (Show, Eq)

newtype InjetiveGettable a b = InjetiveGettable
                          { gettable :: TVar a
                          }

instance (Injective a b) => HasGetter (InjetiveGettable a b) b where
  get r = liftIO $ do
    (value :: a) <- get (gettable r)
    return $ Data.Types.Injective.to value

instance (Injective b a) => HasSetter (InjetiveGettable a b) b where
  t $= a = liftIO $ do
    let b = Data.Types.Injective.to a
    gettable t $= b

newtype DefaultJSON a = DefaultJSON { unDefaultJSON :: a }

instance (FromJSON a, Default a) => FromJSON (DefaultJSON a) where
    parseJSON v = DefaultJSON <$> (parseJSON v <|> pure def)

data AppRefs = AppRefs
  { twsConnectionRefs :: TWSConnectionRefs
  , currentAccount    :: TVar (Maybe Text)
  , appCharts         :: TVar (HashMap Text (TVar Chart))
  , tickerIdToSymbol  :: TVar (HashMap Int Text)
  }

data App = App
  { appSettings :: !Settings
  , appLogFunc :: !LogFunc
  , appProcessContext :: !ProcessContext
  , appOptions :: !Options
  , appWindow :: !Window
  , appRefs :: !AppRefs
  , appData :: !DataRefs
  -- Add other app-specific configuration information here
  }

instance HasLogFunc App where
  logFuncL = lens appLogFunc (\x y -> x { appLogFunc = y })
instance HasProcessContext App where
  processContextL = lens appProcessContext (\x y -> x { appProcessContext = y })
